/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import java.util.List;
import java.util.Map;
import java.util.Random;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.utils.Constants.RoutingType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Triple;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Set;

/**
 * Given a set of nodes and links, with their capacities, an offered traffic vector, and a set of admissible paths for each demand (given by the k-shortest paths
 * in km for each demand, being k a user-defined parameter), this algorithm uses a simulated annealing metaheuristic to find the non-bifurcated routing solution
 * that minimizes the network congestion. Two solutions are considered neighbors if they differ the routing ofone demand. The number of iterations
 * in the inner (all with the same temperature) and outer loop (decreasing the temperature) are user-defined parameters.
 * The initial temperature is computed such that we accept a solution with a 5% of more congestion with 99% of probability. The last
 * temperature is such that with accept a solution with a 5% of more congestion with a 0.01% of probability. Temperature decreases geometrically, the alpha
 * factor of this progression is computed to match previous parameters.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class FA_SAN_minCongestion implements IAlgorithm
{
	private Set<Long> linkIds;
	private long[] demandIdsVector;
	private Map<Long, Double> u_e, h_d;
	private CandidatePathList cpl;
	
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		final int D = netPlan.getNumberOfDemands();
		final int E = netPlan.getNumberOfLinks();
		if (N == 0 || E == 0 || D == 0) throw new Net2PlanException("This algorithm requires a topology with links and a demand set");

		/* Initialize some variables */
		linkIds = netPlan.getLinkIds();
		demandIdsVector = netPlan.getDemandIdsVector();
		u_e = netPlan.getLinkCapacityMap();
		h_d = netPlan.getDemandOfferedTrafficMap();
		
		final int k = Integer.parseInt(algorithmParameters.get("k")); /* number of loopless candidate paths per demand */
		final int san_numOuterIterations = Integer.parseInt(algorithmParameters.get("san_numOuterIterations"));
		final int san_numInnerIterations = Integer.parseInt(algorithmParameters.get("san_numInnerIterations"));

		/* Initial temperature so that we accept bad solution with extra congestion 0.05 with probability 0.99 */
		final double san_initialTemperature = -0.05 / Math.log(0.99);

		/* Final temperature so that we accept bad solution with extra congestion 0.05 with probability 0.001 */
		final double san_finalTemperature = -0.05 / Math.log(0.001);

		/* Alpha factor so that starting in the initial temp, after exactly san_numOuterIterations we have a temperature equal to the final temp  */
		/* In each outer iteration, the new temperature t(i+1) = alpha*t(i) */
		final double san_alpha = Math.pow(san_finalTemperature / san_initialTemperature, 1.0 / (san_numOuterIterations - 1));

		System.out.println("Initial temp: " + san_initialTemperature + ", end temp: " + san_finalTemperature + ", alpha: " + san_alpha);

		/* Construct the candidate path list */
		cpl = new CandidatePathList(netPlan, netPlan.getLinkLengthInKmMap(), "K", Integer.toString(k));

		/* Construct an initial solution: each demand is carried through the shortest path in km  */
		Map<Long, Integer> numberOfDifferentPaths_d = new LinkedHashMap<Long, Integer>();
		Map<Long, Long> current_usedPath_d = new LinkedHashMap<Long, Long>();
		for (long demandId : demandIdsVector)
		{
			Set<Long> pathIds_thisDemand = cpl.getDemandPaths(demandId);
			numberOfDifferentPaths_d.put(demandId, pathIds_thisDemand.size());
			current_usedPath_d.put(demandId, pathIds_thisDemand.iterator().next());
		}
		
		double current_congestion = computeNetCongestion(current_usedPath_d);

		/* Update the best solution found so far to the initial solution */
		Map<Long, Long> best_usedPath_d = new LinkedHashMap<Long, Long>(current_usedPath_d);
		double best_congestion = current_congestion;

		/* Main algorithm loop */
		Random rnd = new Random(); /* random number generator */
		double temp = san_initialTemperature;
		for (int outerItCount = 0; outerItCount < san_numOuterIterations; outerItCount++)
		{
			for (int innerItCount = 0; innerItCount < san_numInnerIterations; innerItCount++)
			{
				/* choose a demand randomly */
				final long demandId = demandIdsVector[rnd.nextInt(D)];
				final int numberPaths = numberOfDifferentPaths_d.get(demandId);
				final long currentPath = current_usedPath_d.get(demandId);

				/* choose a new (different) path for that demand randomly */
				long[] pathIds_thisDemand = cpl.getDemandPathsVector(demandId);
				long newUsedPath = pathIds_thisDemand[rnd.nextInt(numberPaths)];
				while (newUsedPath == currentPath) newUsedPath = pathIds_thisDemand[rnd.nextInt(numberPaths)];
				
				Map<Long, Long> potential_usedPath_d = new LinkedHashMap<Long, Long>(current_usedPath_d);
				potential_usedPath_d.put(demandId, newUsedPath);

				/* compute the new solution cost */
				double newCongestion = computeNetCongestion(potential_usedPath_d);

				/* Check if we accept the new solution */
				if (acceptSolution(current_congestion, newCongestion, temp, rnd))
				{
					current_usedPath_d.put(demandId, newUsedPath);
					current_congestion = newCongestion;

					/* update the best solution if needed */
					if (current_congestion < best_congestion)
					{
						best_congestion = current_congestion;
						best_usedPath_d.clear();
						best_usedPath_d.putAll(current_usedPath_d);
					}
				}
			}
			
			temp *= san_alpha;
		}

		/* Save the best solution found */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(RoutingType.SOURCE_ROUTING);
		for (long demandId : demandIdsVector)
			netPlan.addRoute(demandId, h_d.get(demandId), cpl.getPathSequenceOfLinks(best_usedPath_d.get(demandId)), null);

		return "Ok! Congestion : " + best_congestion;
	}

	@Override
	public String getDescription()
	{
		return "Given a set of nodes and links, with their capacities, an offered traffic vector, and a set of admissible paths for each demand (given by "
				+ "the k-shortest paths in km for each demand, being k a user-defined parameter), this algorithm uses a simulated annealing metaheuristic"
				+ " to find the non-bifurcated routing solution that minimizes the network congestion."
				+ "Two solutions are considered neighbors if they differ the routing ofone demand. The number of iterations in the inner (all with the"
				+ " same temperature) and outer loop (decreasing the temperature) are user-defined parameters. The initial temperature is computed such "
				+ "that we accept a solution with a 5% of more congestion with 99% of probability. The last temperature is such that with accept a "
				+ "solution with a 5% of more congestion with a 0.01% of probability. Temperature decreases geometrically, the alpha factor of this"
				+ " progression is computed to match previous parameters.";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("k", "100", "Number of candidate paths per demand"));
		algorithmParameters.add(Triple.of("san_numOuterIterations", "50", "Number of iterations in the outer loop (changing the temperature)"));
		algorithmParameters.add(Triple.of("san_numInnerIterations", "1000", "Number of iterations in the inner loop (all with the same temperature)"));
		return algorithmParameters;
	}

	/*
	 * SAN criteria to accept a solution according to its cost (to minimize)
	 */
	private static boolean acceptSolution(double current_cost, double new_cost, double temperature, Random rnd)
	{
		if (new_cost < current_cost) return true;

		final double probThreshold = Math.exp(-(new_cost - current_cost) / temperature);
		return rnd.nextDouble() < probThreshold;
	}
	
	/*
	 * Computes the congestion in the network (the objective function of the problem)
	 */
	private double computeNetCongestion(Map<Long, Long> pathId_d)
	{
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);

		for (long demandId : demandIdsVector)
		{
			double h_d_thisDemand = h_d.get(demandId);
			long pathId = pathId_d.get(demandId);
			List<Long> seqLinks = cpl.getPathSequenceOfLinks(pathId);
			for (long linkId : seqLinks) y_e.put(linkId, y_e.get(linkId) + h_d_thisDemand);
		}
		
		return DoubleUtils.maxValue(DoubleUtils.divide(y_e, u_e));
	}
}
